@@ -1,8 +1,15 @@
 import time
 import os
 import json
+import requests as r
+from io import StringIO
+import pandas as pd
+import numpy as np
+import time,sys,csv
 
 class B2bIo:
+  def __init__(self, gapCode='-'):
+      self.gapCode = gapCode
 
   references = [] # Defined in subclass for 'traditional' dynaMine file format output
 
@@ -10,7 +17,7 @@ class B2bIo:
   # File reading
   #
 
-  def readFasta(self, fileName):
+  def readFasta(self, fileName, fileString=None, short_id=True):
 
     """
     Reads a FASTA file - from dynaMine.predictor.parsers import DynaMineFileParsers
@@ -19,31 +26,43 @@ class B2bIo:
     :return: List of (seqId,sequenceString) tuples
     """
 
-    print("Reading FASTA file {}...".format(fileName))
+    if fileName:
+      print("Reading sequences from FASTA file {} ...".format(fileName))
 
-    # Bypassing FC; much faster to do like this...
-    fin = open(fileName)
-    lines = fin.readlines()
-    fin.close()
+      # Bypassing FC; much faster to do like this...
+      with open(fileName, 'r') as fin:
+        lines = fin.readlines()
+
+    else:
+      print("Reading sequences from STRING ...")
+      lines = fileString.split("\n") # WV this might not always work
 
     # Quick FASTA read, handles multiple line sequences
-    seqs = []
+    id_sequence_tuples = []
     seqId = None
     sequence = ""
+
     for line in lines:
       if line.startswith(">"):
         # Consolidate
         if seqId:
-          seqs.append((seqId, sequence))
-        cols = line.split()
-        seqId = self.convertSeqId(cols[0][1:])
+          id_sequence_tuples.append((seqId, sequence.upper()))
+
+        if short_id:
+          cols = line.split()
+          seqId = self.convertSeqId(cols[0][1:])
+        else:
+          seqId = "_".join(line[1:].split()).replace(".", "").replace("|", "_").replace(",", "_")
+
         sequence = ""
       elif line.strip():
+        # RESIDUES
         sequence += line.strip().replace(' ', '')
-    if seqId:
-      seqs.append((seqId, sequence))
 
-    return seqs
+    if seqId:
+      id_sequence_tuples.append((seqId, sequence.upper()))
+
+    return id_sequence_tuples
 
   def writeFasta(self, fastaFileName, sequences):
 
@@ -55,11 +74,10 @@ class B2bIo:
 
     fout.close()
 
-  def convertSeqId(self,seqId):
-
+  def convertSeqId(self, seqId):
     # Possibility to subclass for nicer identifiers (e.g. uniprot default ones)
 
-    return seqId
+    return seqId.replace(' ', '_')
 
   #
   # Wrapper defs for predictions
@@ -202,7 +220,7 @@ class B2bIo:
       elif lines[0].startswith("# STOCKHOLM"):
         fileType = 'STOCKHOLM'
       else:
-        fastaCount = balibaseCount = psiCount = blastCount = emptyLine = 0
+        fastaCount = balibaseCount = psiCount = blastCount = emptyLine = fastaGaps = 0
         for line in lines:
           if line.startswith(">"):
             fastaCount += 1
@@ -214,11 +232,13 @@ class B2bIo:
             blastCount += 1
           elif not line.strip():
             emptyLine += 1
+          elif fastaCount and line.count(gapCode):
+            fastaGaps += 1
 
         if fastaCount > balibaseCount and fastaCount > blastCount:
           if fileName.endswith('a3m'):
             fileType = 'A3M'
-          else:
+          elif fastaGaps:
             fileType = 'FASTA'
         elif blastCount:
           fileType = 'Blast'
@@ -320,7 +340,7 @@ class B2bIo:
     seqIndexes = {}
     seqAlignments = {}
 
-    seqIds = sequences.keys()
+    seqIds = [*sequences.keys()]
     numSeqs = len(seqIds)
     for seqId in seqIds:
       seqIndexes[seqId] = 0
@@ -405,7 +425,7 @@ class B2bIo:
 
     seqAlignments = {}
 
-    seqIds = sequences.keys()
+    seqIds = [*sequences.keys()]
     numSeqs = len(seqIds)
     for seqId in seqIds:
       seqIndexes[seqId] = 0
@@ -556,21 +576,24 @@ class B2bIo:
       if not line.strip():
         continue
 
-      (seqId, sequence) = line.split()
+      try:
+        (seqId, sequence) = line.split()
 
-      seqId = self.getSeqIdKey(seqId)
+        seqId = self.getSeqIdKey(seqId)
 
-      if not self.alignRefSeqID:
-        self.alignRefSeqID = seqId
+        if not self.alignRefSeqID:
+          self.alignRefSeqID = seqId
 
-      # For some reason additional - at beginning of sequence for target protein
-      if i == 0 and sequence[0] == self.gapCode:
-        sequence = sequence[1:]
+        # For some reason additional - at beginning of sequence for target protein
+        if i == 0 and sequence[0] == self.gapCode:
+          sequence = sequence[1:]
 
-      if seqId not in seqAlignments.keys():
-        seqAlignments[seqId] = ""
+        if seqId not in seqAlignments.keys():
+          seqAlignments[seqId] = ""
 
-      seqAlignments[seqId] += sequence
+        seqAlignments[seqId] += sequence
+      except ValueError:
+        continue
 
     return seqAlignments
 
@@ -584,7 +607,7 @@ class B2bIo:
 
     seqCounterToId = {}
     seqCounter = 0
-    seqStartIndex = 51
+    seqStartIndex = 11
 
     fileInfo = lines[0].split()
     (totalSeqs, seqAlignLen) = [int(count) for count in fileInfo]
@@ -598,8 +621,8 @@ class B2bIo:
         seqStartIndex = 0
 
       else:
-
-        sequence = line[seqStartIndex:].replace(" ", "")
+        seqStartIndex = line.rfind(" ") + 1
+        sequence = line[seqStartIndex:] # .replace(" ", "")
 
         if i <= totalSeqs:
           seqId = self.getSeqIdKey(line[:seqStartIndex].strip())
@@ -672,7 +695,7 @@ class B2bIo:
     # Writes a FASTA file from an alignment
     #
 
-    seqIds = seqAlignments.keys()
+    seqIds = [*seqAlignments.keys()]
     seqIds.sort()
 
     fout = open(outFastaFile, 'w')
@@ -685,6 +708,25 @@ class B2bIo:
   # JSON dumps
   #
 
+  def retrieve_seq_uniprot(uniprotID):
+
+    baseUrl = "http://www.uniprot.org/uniprot/"
+    currentUrl = baseUrl + uniprotID + ".fasta"
+    response = r.post(currentUrl)
+    cData = ''.join(response.text)
+
+    Seq = StringIO(cData)
+    pSeq = self.readFasta(None,fileString=Seq)
+    sequence = [(uniprotID, ''.join([res for res in pSeq[0]]))]
+
+    return(sequence)
+
+  def round_floats(self, o):
+    if isinstance(o, float): return round(o, 3)
+    if isinstance(o, dict): return {k: self.round_floats(v) for k, v in o.items()}
+    if isinstance(o, (list, tuple)): return [self.round_floats(x) for x in o]
+    return o
+
   def getAllPredictionsJson(self, identifier, limitPrecision=True):
 
     """
@@ -719,7 +761,76 @@ class B2bIo:
     if limitPrecision:
       json.encoder.FLOAT_REPR = lambda o: format(o, '.3f')
 
-    return(json.dumps(jsonData, sort_keys=True, indent=4, separators=(',', ': ')))
+
+    # return(json.dumps(jsonData, sort_keys=True, indent=4, separators=(',', ': ')))
+    return json.dumps(self.round_floats(jsonData), sort_keys=True, indent=4, separators=(',', ': '))
+
+  def getAllPredictionsJson_msa(self, results, limitPrecision=True):
+
+      """
+      Creates JSON string with all available predictions with identifier as base name
+      :param identifier: Base name for the output files
+      :return: JSON with full information for all predictions done in the subclass
+      """
+
+      jsonData = {
+        'creation_date': time.strftime("%Y-%m-%dT%H:%M:%S"),
+        'information': self.informationPerPredictor,
+        'results': results
+      }
+
+      json_data = json.dumps(self.round_floats(jsonData), sort_keys=True, indent=4, separators=(',', ': '))
+      return jsonData
+
+  def json_to_fasta_variants(json_filename, output_dir=""):
+    with open(json_filename) as json_file:
+      data = json.load(json_file)
+      # Generate fasta file with the WT & variants
+      filename = os.path.join(output_dir, "{}.fasta".format(data["metadata"]["name"]))
+      fasta_file = open(filename, "w")
+      # Add the WT
+      fasta_file.write(">WT\n")
+      fasta_file.write(data["WT"] + "\n")
+      # Add the variants
+      for key in data["Variants"].keys():
+        fasta_file.write(">{}\n".format(key))
+        # Add all the mutations per variant
+        mut_seq = list(data["WT"])
+        adp_val = 0 # Value to adapt the numbering affected by insertions
+        for mut in data["Variants"][key]:
+          # Deletions
+          if "del" in mut:
+            pos = int(mut[1:-3]) - 1 + adp_val
+            mut_seq[pos] = "-"
+
+          # Insertions
+          elif "ins" in mut:
+            pos = int(mut.split("_")[0][1:]) + adp_val
+
+            for aa in mut.split("ins")[1]:
+              mut_seq.insert(pos + adp_val, aa)
+              adp_val += 1
+
+          else:
+            # Identify the position to be mutated
+            pos = int(mut[1:-1]) - 1 + adp_val
+            # Mutate the residue
+            mut_seq[pos] = mut[-1]
+
+        # Remove deletion "-" sings if any
+        mut_seq = [i for i in mut_seq if i != "-"]
+
+        # Add the mutated sequences in the fasta file
+        fasta_file.write("{}\n".format("".join(mut_seq)))
+
+    fasta_file.close()
+    # Run the MSA using all the BLAST matches found
+    aligned_file = '{}_tcof.aln'.format(filename.split('.')[0])
+    cmd = "t_coffee {} -output=fasta_aln -outfile={}".format(
+      filename, aligned_file)
+
+    os.system(cmd)
+    return aligned_file
 
   #
   # NEF file handling
@@ -736,7 +847,7 @@ class B2bIo:
 
     # Note that this nmrStar reader is generic STAR, not specifically for NMR-STAR.
     # If this works, could also be an NMR-STAR file!
-    origNefFile = starReader.File(verbosity=2, filename=fileName)
+    origNefFile = starReader.File(verbosity=0, filename=fileName)
 
     #
     # Read NEF file (in STAR format)
@@ -745,6 +856,10 @@ class B2bIo:
     if origNefFile.read():
       print("  Error reading NEF file ")
       return None
+    else:
+      if not origNefFile.datanodes:
+        print("  No NEF data available.")
+        return None
 
     return origNefFile
 
@@ -1037,3 +1152,95 @@ class B2bIo:
                                                                                        chemShift.atomName,
                                                                                        chemShift.value,
                                                                                        chemShift.valueError))
+
+  def json_preds_to_csv_singleseq(self, json_data, output_dir="results"):
+      # Function to save the predictions in the json file to a txt-like file
+      filename = json_data["id"].split("/")[-1].split(".")[0]
+      output_file = os.path.join(output_dir, '{}_b2btools_preds.txt'.format(filename))
+
+      with open(output_file, 'w+') as predfile:		# outfile here
+        writer = csv.writer(predfile, delimiter=',')
+
+        headerWritten = False
+        for preds in json_data["results"]:  # iterate the predictions and get information to write to csv
+          seqid = preds["proteinID"]
+
+          predNames = ('sequence', 'backbone', 'sidechain', 'sheet', 'helix', 'coil', 'earlyFolding', 'disoMine', 'agmata')
+          predNamesFound = []
+          predsFound = []
+
+          for predName in predNames:
+            if predName in preds.keys():
+              predNamesFound.append(predName)
+
+              if predName in ('sequence'):
+                predsFound.append([elem for elem in preds[predName]])
+              else:
+                predsFound.append([getCsvString(elem) for elem in preds[predName]])
+
+          # WV to fix!
+          # Fix sequence to alignment, if this is a msa-connected prediction
+          #if tool == 'msatoolsproteins':
+          #  seqNumberCode = 'AlignCode'
+          #  for i in range(len(predsFound[1])):
+              # If None, is a gap, so insert None in sequence as well
+          #    if predsFound[1][i] == 'None':
+          #      predsFound[0].insert(i, '-')
+          #else:
+          seqNumberCode = 'SeqCode'
+
+          seqLen = len(predsFound[0])  # Sequence length
+
+          # Insert sequence numbers
+          predNamesFound.insert(0, seqNumberCode)
+          predsFound.insert(0, range(1, seqLen + 1))
+
+          # Insert protein info
+          predNamesFound.insert(0, "Accession")
+          predsFound.insert(0, [seqid] * seqLen)
+
+          if not headerWritten:
+            writer.writerow(predNamesFound)
+            headerWritten = True
+
+          allpredsperseq = zip(*predsFound)
+          for row in allpredsperseq:
+            writer.writerow(row)
+
+  def json_preds_to_csv_msa(self, json_data, id, output_dir = 'results'):
+        # Function to save the predictions in the json file to a txt-like file
+        output_file = os.path.join(output_dir, '{}_b2btools_preds.txt'.format(id))
+
+        with open(output_file, 'w+') as predfile:		# outfile here
+          preds = json_data["results"]["results"]
+          #preds = response_dict['statistic']
+
+          writer = csv.writer(predfile, delimiter=',')
+
+          predCodes = list(preds.keys())
+          predCodes.sort()
+
+          predNames = ('backbone', 'sidechain', 'sheet', 'helix', 'coil', 'earlyFolding', 'disoMine', 'agmata')
+          statsNames = ("median", "topOutlier", "firstQuartile", "thirdQuartile", "bottomOutlier")
+
+          predNamesFound = []
+          predsFound = []
+
+          for predName in predNames:
+            if predName in preds.keys():
+              for statsName in statsNames:
+                if statsName in preds[predName].keys():
+                  predNamesFound.append("{}_{}".format(predName, statsName))
+                  predsFound.append([getCsvString(elem) for elem in preds[predName][statsName]])
+
+          alignLen = len(predsFound[0])  # Alignment length
+
+          # Insert alignment numbers
+          predNamesFound.insert(0, 'AlignCode')
+          predsFound.insert(0, range(1, alignLen + 1))
+
+          writer.writerow(predNamesFound)
+
+          allpredsperseq = zip(*predsFound)
+          for row in allpredsperseq:
+            writer.writerow(row)
